import os, random, subprocess, sys, time

from datetime import datetime
from pathlib import Path
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from .config import setup_config, logging


config = setup_config()


class WatchHandler(PatternMatchingEventHandler):
	patterns = [f'*.{ext}' for ext in config.watch_ext]


	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.proc = None
		self.last_restart = None


	def kill_proc(self):
		if self.proc.poll() != None:
			return

		logging.info(f'Terminating process {self.proc.pid}')
		self.proc.terminate()
		sec = 0.0

		while self.proc.poll() == None:
			time.sleep(0.1)
			sec += 0.1

			if sec >= 5:
				logging.error('Failed to terminate. Killing process...')
				self.proc.kill()
				break

		logging.info('Process terminated')


	def run_proc(self, restart=False):
		timestamp = datetime.timestamp(datetime.now())
		self.last_restart = timestamp if not self.last_restart else 0

		if restart == True and self.proc.pid != '':
			if timestamp - 3 < self.last_restart:
				logging.debug('Process restarted less than 1 sec ago')
				return

			self.kill_proc()

		cmd = [config.bin]

		if config.env:
			env = ['env']

			for k,v in config.env.items():
				env.append(f'{k}={v}')

			cmd = [*env, *cmd]

		try:
			self.proc = subprocess.Popen([*cmd, *config.args], preexec_fn=os.setpgrp, stdin=subprocess.PIPE)
		except FileNotFoundError:
			logging.error(f'Cannot find executable: {config.bin}')
			sys.exit()

		self.last_restart = timestamp
		logging.info(f'Started process with PID {self.proc.pid}')


	def on_any_event(self, event):
		if event.event_type not in ['modified', 'created', 'deleted']:
			logging.debug(f'Ignored event: {event.event_type}')
			return

		path = Path(event.src_path)
		directory = str(path.parent).replace(str(config.path), '')

		if not set(config.ignore_dirs).intersection(directory.split('/')):
			for filename in config.ignore_files:
				if path.name == filename:
					logging.debug(f'Ignored file name: {filename}')
					return

			self.run_proc(restart=True)


def main():
	logging.info('Starting process watcher')

	handler = WatchHandler()
	handler.run_proc()

	watcher = Observer()
	watcher.schedule(handler, str(config.path), recursive=True)
	watcher.start()

	timeout = None
	timer = 0

	try:
		while True:
			handler.proc.stdin.write(sys.stdin.read().encode('UTF-8'))
			handler.proc.stdin.flush()

	except KeyboardInterrupt:
		pass

	handler.kill_proc()
	logging.debug('Stopping process watcher')
	watcher.stop()
	watcher.join()
